/*	coreVarsGenerator.c

	Copyright 2008 Fred Cooke

	This file is part of the FreeEMS project.

	FreeEMS software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FreeEMS software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any FreeEMS software.  If not, see <http://www.gnu.org/licenses/>.

	We ask that if you make any changes to this file you send them upstream to us at admin@diyefi.org

	Thank you for choosing FreeEMS to run your engine! */

#define COREVARSGENERATOR_C
#include "inc/freeEMS.h"
#include "inc/commsISRs.h"
#include "inc/coreVarsGenerator.h"


void generateCoreVars(){
	/*&&&&&&&& Calculate and obtain the basic variables with which we will perform the calculations &&&&&&&&*/


	/* Pre calculate things used in multiple places */

	/* Bound the TPS ADC reading and shift it to start at zero */
	unsigned short unboundedTPSADC = ADCArrays.TPS[mathInputBank];
	if(unboundedTPSADC > configs.TPSMaximumADC){
		boundedTPSADC = TPSADCRange;
	}else if(unboundedTPSADC > configs.TPSMinimumADC){
		boundedTPSADC = unboundedTPSADC - configs.TPSMinimumADC;
	}else{
		boundedTPSADC = 0;
	}

	
	/* Get BRV from ADC using transfer variables (all installations need this) */
	unsigned short localBRV;
	if(TRUE){ /* If BRV connected  */
		localBRV = (((unsigned long)ADCArrays.BRV[mathInputBank] * configs.BRVRange) / ADC_DIVISIONS) + configs.BRVMinimum;
	}else if(FALSE){ /* Configured to be fixed value */
		/* Get the preferred BRV figure from configuration settings */
		localBRV = configs.presetBRV;
	}else{ /* Fail safe if config is broken */
		/* Default to normal alternator charging voltage 14.4V */
		localBRV = runningVoltage;
		/* If anyone is listening, let them know something is wrong */
		sendError(BRV_NOT_CONFIGURED_CODE);
	}


	unsigned short localCHT;
	/* Get CHT from ADC using the transfer table (all installations need this) */
	if(TRUE){ /* If CHT connected  */
		localCHT = CHTTransferTable[ADCArrays.CHT[mathInputBank]];
	}else if(FALSE){ /* Configured to be read From ADC as dashpot */
		/* Transfer the ADC reading to an engine temperature in a reasonable way */
		localCHT = (ADCArrays.CHT[mathInputBank] * 10) + freezingPoint; /* 0 ADC = 0C = 273.15K = 27315, 1023 ADC = 102.3C = 375.45K = 37545 */
	}else if(FALSE){ /* Configured to be fixed value */
		/* Get the preferred CHT figure from configuration settings */
		localCHT = configs.presetCHT;
	}else{ /* Fail safe if config is broken */
		/* Default to normal running temperature of 85C/358K */
		localCHT = runningTemperature;
		/* If anyone is listening, let them know something is wrong */
		sendError(CHT_NOT_CONFIGURED_CODE);
	}


	unsigned short localIAT;
	/* Get IAT from ADC using the transfer table (all installations need this) */
	if(TRUE){ /* If IAT connected  */
		localIAT = IATTransferTable[ADCArrays.IAT[mathInputBank]];
	}else if(FALSE){ /* Configured to be read From ADC as dashpot */
		/* Transfer the ADC reading to an air temperature in a reasonable way */
		localIAT = (ADCArrays.IAT[mathInputBank] * 10) + 27315; /* 0 ADC = 0C = 273.15K = 27315, 1023 ADC = 102.3C = 375.45K = 37545 */
	}else if(FALSE){ /* Configured to be fixed value */
		/* Get the preferred IAT figure from configuration settings */
		localIAT = configs.presetIAT;
	}else{ /* Fail safe if config is broken */
		/* Default to normal air temperature of 20C/293K */
		localIAT = roomTemperature;
		/* If anyone is listening, let them know something is wrong */
		sendError(IAT_NOT_CONFIGURED_CODE);
	}


	unsigned short localMAT;
	/* Determine the MAT reading for future calculations */
	if(TRUE){ /* If MAT sensor is connected */
		/* Get MAT from ADC using same transfer table as IAT (too much space to waste on having two) */
		localMAT = IATTransferTable[ADCArrays.MAT[mathInputBank]];
	}else if(FALSE){ /* Configured to be fixed value */
		/* Get the preferred MAT figure from configuration settings */
		localMAT = configs.presetMAT;
	}else{ /* Fail safe if config is broken */
		/* If not, default to same value as IAT */
		localMAT = localIAT;
		/* If anyone is listening, let them know something is wrong */
		sendError(MAT_NOT_CONFIGURED_CODE);
	}

	
	unsigned short localMAP;
	unsigned short localIAP;
	/* Determine the MAP pressure to use for future calculations */
	if(TRUE){ /* If MAP sensor is connected */
		/* get MAP from ADC using transfer variables */
		localMAP = (((unsigned long)ADCArrays.MAP[mathInputBank] * configs.MAPRange) / ADC_DIVISIONS) + configs.MAPMinimum;
		if(TRUE){ /* If Intercooler boost sensor connected */
			/* Get IAP from ADC using the same transfer variables as they both need to read the same range */
			localIAP = (((unsigned long)ADCArrays.IAP[mathInputBank] * configs.MAPRange) / ADC_DIVISIONS) + configs.MAPMinimum;
		}
	}else if(FALSE){ /* Configured for MAP to imitate TPS signal */
		/* Get MAP from TPS via conversion */
		localMAP = (((unsigned long)boundedTPSADC * TPSMAPRange) / TPSADCRange) + configs.TPSClosedMAP;
	}else if(FALSE){ /* Configured for dash potentiometer on ADC */
		/* Get MAP from ADC via conversion to internal kPa figure where 1023ADC = 655kPa */
		localMAP = ADCArrays.MAP[mathInputBank] << 6;
		if(TRUE){ /* If Intercooler boost sensor enabled */
			/* Get IAP from ADC via conversion to internal kPa figure where 1023ADC = 655kPa */
			localIAP = ADCArrays.IAP[mathInputBank] << 6;
		}
	}else if(FALSE){ /* Configured for fixed MAP from config */
		/* Get the preferred MAP figure from configuration settings */
		localMAP = configs.presetMAP;
	}else{ /* Fail safe if config is broken */
		/* Default to zero to nulify all other calcs and effectively cut fuel */
		localMAP = 0;
		/* If anyone is listening, let them know something is wrong */
		sendError(MAP_NOT_CONFIGURED_CODE); // or maybe queue it?
	}


	/* Determine MAF variable if required */
	unsigned short localMAF = 0; // Default to zero as it is not required for anything except main PW calcs optionally
	if(FALSE){
		localMAF = MAFTransferTable[ADCArrays.MAF[mathInputBank]];
	}
	
	unsigned short localAAP;
	/* Determine the Atmospheric pressure to use for future calculations */
	if(TRUE){ /* Configured for second sensor to read AAP */
		/* get AAP from ADC using separate vars to allow 115kPa sensor etc to be used */
		localAAP = (((unsigned long)ADCArrays.AAP[mathInputBank] * configs.AAPRange) / ADC_DIVISIONS) + configs.AAPMinimum;
	}else if(FALSE){ /* Configured for dash potentiometer on ADC */
		/* Get AAP from ADC via conversion to internal kPa figure where 1023ADC = 102.3kPa */
		localAAP = ADCArrays.AAP[mathInputBank] * 10;
	}else if(FALSE){ /* Configured for fixed AAP reading from pre start */
		/* Get the AAP reading as saved during startup */
		localAAP = bootTimeAAP; /* This is populated pre start up */
	}else if(FALSE){ /* Configured for fixed AAP from config */
		/* Get the preferred AAP figure from configuration settings */
		localAAP = configs.presetAAP;
	}else{ /* Fail safe if config is broken */
		/* Default to sea level */
		localAAP = seaLevelKPa; /* 100kPa */
		/* If anyone is listening, let them know something is wrong */
		sendError(AAP_NOT_CONFIGURED_CODE); // or maybe queue it?
	}


	unsigned short localEGO;
	/* Get main Lambda reading */
	if(TRUE){ /* If WBO2-1 is connected */
		/* Get EGO from ADCs using transfer variables */
		localEGO = (((unsigned long)ADCArrays.EGO[mathInputBank] * configs.EGORange) / ADC_DIVISIONS) + configs.EGOMinimum;
	}else if(FALSE){ /* Configured for fixed EGO from config */
		/* Get the preferred EGO figure from configuration settings */
		localEGO = configs.presetEGO;
	}else{ /* Default value if not connected incase other things are misconfigured */
		/* Default to stoichiometric */
		localEGO = stoichiometricLambda; /* EGO / 32768 = Lambda */
		/* If anyone is listening, let them know something is wrong */
		sendError(EGO_NOT_CONFIGURED_CODE); // or maybe queue it?
	}


	unsigned short localEGO2;
	/* Get second Lambda reading */
	if(TRUE){ /* If WBO2-2 is connected */
		/* Get EGO2 from ADCs using same transfer variables as EGO */
		localEGO2 = (((unsigned long)ADCArrays.EGO2[mathInputBank] * configs.EGORange) / ADC_DIVISIONS) + configs.EGOMinimum;
	}else if(FALSE){ /* Configured for fixed EGO2 from config */
		/* Get the preferred EGO2 figure from configuration settings */
		localEGO2 = configs.presetEGO2;
	}else{ /* Default value if not connected incase other things are misconfigured */
		/* Default to stoichiometric */
		localEGO2 = stoichiometricLambda;
		/* If anyone is listening, let them know something is wrong */
		sendError(EGO2_NOT_CONFIGURED_CODE); // or maybe queue it?
	}


	unsigned short localTPS;
	/* Get TPS percentage */
	if(TRUE){ /* If TPS is connected */
		/* Get TPS from ADC no need to add TPS min as we know it is zero by definition */
		localTPS = ((unsigned long)boundedTPSADC * TPS_RANGE_MAX) / TPSADCRange;
	}else if(FALSE){ /* Configured for TPS to imitate MAP signal */
		/* Get TPS from MAP via conversion */
		/* Box MAP signal down */
		if(localTPS > configs.TPSOpenMAP){ /* Greater than ~95kPa */
			localTPS = TPS_RANGE_MAX; /* 64000/640 = 100% */
		}else if(localTPS < configs.TPSClosedMAP){ /* Less than ~30kPa */
			localTPS = 0;
		}else{ /* Scale MAP range to TPS range */
			localTPS = localMAP - configs.TPSClosedMAP;
		}
		// get TPS from MAP no need to add TPS min as we know it is zero by definition
		localTPS = ((unsigned long)localTPS * TPS_RANGE_MAX) / (configs.TPSOpenMAP - configs.TPSClosedMAP); 
	}else if(FALSE){ /* Configured for dash potentiometer on ADC */
		/* Get TPS from ADC as shown : 1023 ADC = 100%, 0 ADC = 0% */
		localTPS = ((unsigned long)ADCArrays.TPS[mathInputBank] * TPS_RANGE_MAX) / ADC_DIVISIONS;
	}else if(FALSE){ /* Configured for fixed TPS from config */
		/* Get the preferred TPS figure from configuration settings */
		localTPS = configs.presetTPS;
	}else{ /* Fail safe if config is broken */
		/* Default to 50% to not trigger any WOT or CT conditions */
		localTPS = halfThrottle;
		/* If anyone is listening, let them know something is wrong */
		sendError(TPS_NOT_CONFIGURED_CODE); // or maybe queue it?
	}


	/* Get RPM by locking out ISRs for a second and grabing the Tooth logging data */
	//atomic start
	// copy rpm data
	//atomic end
	
	// Calculate RPM and delta RPM and delta delta RPM from data recorded
	CoreVars.RPM[mathInternalBank] = RPM[mathInputBank]; // temporary!!
	unsigned short localDRPM = 0;
	unsigned short localDDRPM = 0;


	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/




	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&& Average the variables as per the configuration &&&&&&&&&&&&&&&&&&&&&&&&&&*/
	/* Strictly speaking only the primary variables need to be averaged. After that, the derived ones are	*/
	/* already averaged in a way. However, there may be some advantage to some short term averaging on the	*/
	/* derived ones also, so it is something to look into later.											*/

	// TODO average them here

//			newVal var word        ' the value from the ADC
//			smoothed var word    ' a nicely smoothed result 
//
//			if newval > smoothed then
//			        smoothed = smoothed + (newval - smoothed)/alpha
//			else
//			        smoothed = smoothed - (smoothed - newval)/alpha
//			endif

	// from : http://www.tigoe.net/pcomp/code/category/code/arduinowiring/41			
	
	// for now just copy them in.
	CoreVars.IAT[mathInternalBank] = localIAT;
	CoreVars.CHT[mathInternalBank] = localCHT;
	CoreVars.TPS[mathInternalBank] = localTPS;
	CoreVars.EGO[mathInternalBank] = localEGO;
	CoreVars.BRV[mathInternalBank] = localBRV;
	CoreVars.MAP[mathInternalBank] = localMAP;
	CoreVars.AAP[mathInternalBank] = localAAP;
	CoreVars.MAT[mathInternalBank] = localMAT;

	CoreVars.EGO2[mathInternalBank] = localEGO2;
	CoreVars.IAP[mathInternalBank] = localIAP;
	CoreVars.MAF[mathInternalBank] = localMAF;
	CoreVars.DRPM[mathInternalBank] = localDRPM;
	CoreVars.DDRPM[mathInternalBank] = localDDRPM;
	
	// later...
	unsigned short i;
	for(i=0;i<CORE_VARS_LENGTH;i++){ // TODO
		/* Perform averaging on all primary variables as per the configuration array */
		// get old value(s)
		// process new and old to produce result based on config array value */
		// assign result to old value holder
	} // TODO

	/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
}
