/*	flashWrite.c

	Copyright 2008 Sean Keys

	This file is part of the FreeEMS project.

	FreeEMS software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FreeEMS software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any FreeEMS software.  If not, see <http://www.gnu.org/licenses/>.

	We ask that if you make any changes to this file you send them upstream to us at admin@diyefi.org

	Thank you for choosing FreeEMS to run your engine! */

#define FLASHWRITE_C
#include "inc/freeEMS.h"
#include "inc/flashWrite.h"
#include "inc/flashBurn.h"
#include "inc/commsISRs.h"


/**************************************************************************************************************/
/*	27.4.2.4 Sector Erase Command

	The sector erase operation will erase all addresses in a 1 Kbyte sector of Flash memory using an embedded
	algorithm.

	An example flow to execute the sector erase operation is shown in Figure 27-29. The sector erase
	command write sequence is as follows:

	1. Write to a Flash block address to start the command write sequence for the sector erase command.
	The Flash address written determines the sector to be erased while global address bits [9:0] and the
	data written are ignored. Multiple Flash sectors can be simultaneously erased by writing to the
	same relative address in each Flash block.

	2. Write the sector erase command, 0x40, to the FCMD register.

	3. Clear the CBEIF flag in the FSTAT register by writing a 1 to CBEIF to launch the sector erase
	command.

	If a Flash sector to be erased is in a protected area of the Flash block, the PVIOL flag in the FSTAT register
	will set and the sector erase command will not launch. Once the sector erase command has successfully
	launched, the CCIF flag in the FSTAT register will set after the sector erase operation has completed unless
	a new command write sequence has been buffered. */
unsigned char eraseSector(unsigned short *flash_address){
	unsigned short FlashSectorSize = 1024;

	/* register D will contain the address passed into eraseSector aka flash_address */
	if ((unsigned short)flash_address & 0x001){
			send("Flash_Address_Not_Square  ");
			sendUS((short)flash_address);
		return Flash_Address_Not_Aligned;
	}

	if ((unsigned short)flash_address % FlashSectorSize != 0){
		send("Not_FlashSectorSize  ");
		sendUS((unsigned short)flash_address);
		send(" % ");
		sendUS((unsigned short)FlashSectorSize);
		send("=");
		unsigned short modulo= (unsigned short)flash_address % FlashSectorSize;
		sendUS(modulo);
		return Address_Not_Start_Of_Flash_Sector;
		}

	FSTAT = (PVIOL|ACCERR); /* clear any errors */
	(*flash_address) = 0xFFFF;     /* Dummy store to page to be erased */
	FCMD = SECTOR_ERASE;            /* set the flash command register mode to ERASE */
	send(", Calling StackBurner");
	StackBurner(flash_address);  //PPAGE loaded into Register B, PPAGE to be implemented

	//TODO add return for accerr and pviol error bits

	return Command_Successful;
}


/*******************************************************************************
 * 	writeBlock will use writeWord to write a 1k block from RAM to flash.
 *  Give it the starting memory address and the destination flash address.
 *  Both addresses will be incremented by 1 word after a successful writeWord,
 *  until the whole 1024 byte sector has been written.  Before any writing occurs
 *  eraseSector is called to make sure the destination is blank. */
unsigned char writeAlignedBlock(unsigned short *flashAddr, unsigned short *RAMAddr){

	if ((unsigned short)flashAddr % FlashSectorSize != 0){
		send("Not_FlashSectorSize  ");
		sendUS((unsigned short)flashAddr);
		return Address_Not_Start_Of_Flash_Sector;
	}

	eraseSector((unsigned short*)flashAddr);  /* First Erase our destination block */
	unsigned short wordCount = 512;  /* 512 words to a 1024k sector */
	while (wordCount > 0)
	{
		unsigned short data = *RAMAddr; /*Convert the RAMAddr to data(dereference) */
		writeWord((unsigned short*)flashAddr, data);
		*RAMAddr = *RAMAddr + 2; /* increment RAM source up by one word */
		*flashAddr = *flashAddr + 2; /* increment flash destination up by one word */
		wordCount--; /* Decrement our word counter */
	}

	return 1;
}


/*	27.4.2.3 Program Command

	The program operation will program a previously erased word in the Flash memory using an embedded
	algorithm.

	An example flow to execute the program operation is shown in Figure 27-28. The program command write
	sequence is as follows:

	1. Write to a Flash block address to start the command write sequence for the program command. The
	data written will be programmed to the address written. Multiple Flash blocks can be
	simultaneously programmed by writing to the same relative address in each Flash block.

	2. Write the program command, 0x20, to the FCMD register.

	3. Clear the CBEIF flag in the FSTAT register by writing a 1 to CBEIF to launch the program
	command.

	If a word to be programmed is in a protected area of the Flash block, the PVIOL flag in the FSTAT register
	will set and the program command will not launch. Once the program command has successfully launched,
	the CCIF flag in the FSTAT register will set after the program operation has completed unless a new
	command write sequence has been buffered. By executing a new program command write sequence on
	sequential words after the CBEIF flag in the FSTAT register has been set, up to 55% faster programming
	time per word can be effectively achieved than by waiting for the CCIF flag to set after each program
	operation.*/
unsigned char writeWord(unsigned short *flash_address, unsigned short data){
    // TODO Create error check methods to get rid of redundant code

	if ((unsigned short)flash_address & 0x0001){
		send("Flash_Address_Not_Square  ");
		sendUS((unsigned short)flash_address);
		return Flash_Address_Not_Aligned;
	}

	FSTAT=(ACCERR | PVIOL);
    (*flash_address)= data;
    FCMD = WORD_PROGRAM;        //Load Flash Command Register With Word_Program mask
    send(", Calling StackBurner");
    StackBurner(flash_address); //PPAGE loaded into Register B, PPAGE to be implemented

    return Command_Successful;
}
